
{-
Copyright 2009, Jitesh Jain, Shashank Kapoor, Sajal Kumar Sinha, Piyush Srivastava and Utkarsh Upadhyay.

This file is part of atp, an automated theorem prover.

atp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

atp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with atp.  If not, see <http://www.gnu.org/licenses/>.

-}

{-# OPTIONS_GHC -cpp #-}
module Resolver1 where
import Control.Concurrent
import Control.Exception as Exception
import Control.Monad
import Data.List as List
import Data.Sequence as Seq
import Logic1 as Logic
import System.Console.Readline
import System.IO
import Text.ParserCombinators.Parsec 
import Text.ParserCombinators.Parsec.Token as Parsec
import Text.ParserCombinators.Parsec.Language
import Unification1 as Unification
import qualified Data.Foldable as Foldable
import qualified Data.Map as Map
import qualified Resolution1 as Resolution

type IndexedClause = (Maybe (Int, Int), Resolution.IndexedClause)

--Parsers for IO
languageDef :: LanguageDef ()
languageDef = haskellDef

lexan :: Parsec.TokenParser ()
lexan = Parsec.makeTokenParser haskellDef
lexeme = Parsec.lexeme lexan

comma = Parsec.comma lexan
integer = Parsec.integer lexan
squares = Parsec.squares lexan
whiteSpace = Parsec.whiteSpace lexan

integer1 = Resolver1.lexeme (do{Resolver1.whiteSpace; Resolver1.integer;})

int = do { x<- Resolver1.integer; return $ ((fromIntegral x)::Int)}
intList = Resolver1.lexeme (Resolver1.whiteSpace >> (Resolver1.squares (sepBy Resolver1.int Resolver1.comma)))

{- |Checks if the thing returned by lrIndexedResolve is already a refutation-}
done :: [(Maybe (Int, Int), CnfFormula)] -> Bool
done ((_,[]):_) = True
done _ = False


lrIndexedResolve :: IndexedClause -> IndexedClause -> [(Maybe (Int, Int), CnfFormula)]
lrIndexedResolve  (p, (n1, f11)) (q, (n2, f12)) = let  (f1, f2) = Resolution.modifyPair f11 f12
                                                        in map (\x-> (Just(n1,n2), x)) (lrResolve f1 f2)
lrResolve :: CnfFormula -> CnfFormula -> [CnfFormula]
lrResolve f1 f2 = let c1 = Resolution.merge f1 f2
                      list = Resolution.isInOrderForSup f1 f2
                      lrResolveStep ((l1,l2):xs) i = let p = Unification.unifyAFs (map Resolution.stripAF [l1,l2])
                                                         c2 = c1 \\ [l1, l2]
                                                         in case p of 
                                                               Right _-> lrResolveStep xs i
                                                               Left s -> let temp = factor $ snd $
                                                                                Resolution.modify 0
                                                                                 (Resolution.substitute s c2)
                                                                             in if temp == [] 
                                                                                   then ([]:i)
                                                                                   else lrResolveStep xs (temp:i)
                      lrResolveStep [] i = i
                      in (List.nub $ lrResolveStep list [])
                      

sameFunc :: AtomicFormula -> AtomicFormula -> Bool
sameFunc x y = (afName x == afName y) && (List.length (afArg x) == List.length (afArg y))

unifyLitPair :: (Literal, Literal)  -> Bindings
unifyLitPair (Not x, Not y) = if sameFunc x y then Unification.unifyAFs [x,y] else Right "err"
unifyLitPair (Pos x, Pos y) = if sameFunc x y then Unification.unifyAFs [x,y] else Right "err"
unifyLitPair (_, _) = Right "err" 


factor :: CnfFormula -> CnfFormula
factor x = let p = [(a,b)| a <- x, b <- x]
               in nub (factor_ p x)

factor_ [] x = x 
factor_ (p@(a,b):ps) x = let c = unifyLitPair (a,b)
                               in case c of
                                       Right _ -> factor_ ps x
                                       Left s  -> factor_ ps (Resolution.substitute s x)
        





{-
#ifdef DEBUG
-}
showIndexed :: IndexedClause -> String
showIndexed = show


showIndexedList :: [IndexedClause] -> String
showIndexedList = show

{-
#else
-}
showIndexed :: IndexedClause -> String
showIndexed (Just (a,b),(i, c)) = show a ++ ", " ++ show b ++ " -> " ++ show i ++ ": " ++ Logic.showCnf c
showIndexed (Nothing,(i, c)) = show i ++ ": " ++ Logic.showCnf c

showIndexedList :: [IndexedClause] -> String
showIndexedList l = concat $ List.intersperse "\n" (List.map showIndexed l)

{-
#endif
-}
{- | This takes care of the actual resolution. The state must be initialized to
- a list of indexed input clauses.From a set of support and the database, it finds the next set of support.-}

resolver :: [IndexedClause] -> [IndexedClause] -> Int -> (Int, Bool, [IndexedClause])
resolver support db n = let --d = indexedResolveNub (do {x <- support; y <- db; lrIndexedResolve x y })
                            d1 = [(x,y) | x <- support, y <- db]
                            repeatIndexedResolve ((a,b):xs) = let p = lrIndexedResolve a b
                                                                  q = List.reverse p
                                                                  in if done p 
                                                                      then q
                                                                      else q ++ repeatIndexedResolve xs
                            repeatIndexedResolve [] = []
                            d = indexedResolveNub (repeatIndexedResolve d1)
                            c = diff d support
                            ((n1, tf), l) = List.mapAccumL compactify (n, False) c
                            in (n1, tf, l)

{- |To remove elements from a list of the form [(Maybe b, a)], conditioned on the appearance of
- a in a list of the form [(Maybe b, (c,a))]  conditioned on a.-}

diff :: (Eq a) => [(Maybe b, a)] -> [(Maybe b, (c,a))] -> [(Maybe b, a)]
diff = foldl diff'

diff' :: (Eq a) => [(Maybe b, a)] -> (Maybe b, (c,a)) -> [(Maybe b, a)]
diff' [] _ = []
diff' (t@(x,p):xs) s@(y,(q,r)) = if p == r then xs else (t:(diff' xs s))


{- |To remove duplicates from [(Maybe b, a)], conditioned on a.-}
indexedResolveNub :: (Eq a) => [(Maybe b, a)] -> [(Maybe b, a)]
indexedResolveNub [] = []
indexedResolveNub (s@(y,x):xs) = s : (indexedResolveNub $ filter (\(a, b) ->  b /= x) xs)


{- |To remove duplicates from [(Maybe b, (a,c))], conditioned on c.-}
indexedResolveNub' :: (Eq c) => [(Maybe b, (a, c))] -> [(Maybe b, (a,c))]
indexedResolveNub' [] = []
indexedResolveNub' (s@(y,(x', x)):xs) = s : (indexedResolveNub' $ filter (\(a, (p, b)) ->  b /= x) xs)

repeatedResolver :: Int -> [IndexedClause] -> [IndexedClause] -> IO ()
repeatedResolver i support db = do let (a, b, c) = resolver support db i
                                   --putStrLn "\nNew Support: \n"
                                   putStrLn $ "\nNew support: \n" ++  showIndexedList c
                                   --mapM_ print c
                                   if b
                                      then putStrLn "\nProved."
                                      else if c == [] 
                                              then putStrLn "\nCan't be Proved, or db inconsistent."
                                              else repeatedResolver a c (indexedResolveNub' (support ++ db))
                                                       

compactify :: Eq a => (Int, Bool) -> (Maybe b, [a]) -> ((Int, Bool), (Maybe b, (Int, [a])))
compactify (a, b) (y,x) = ((a+1,b || (x == [])), (y, (a, x)))

{- |Converts a list of Cnfs to a list of IndexedClause. To be called with
- first argument as 0-}                             
toIndexedCnf :: Int -> [a] -> [(Maybe (Int,Int), (Int,a))]
toIndexedCnf _ [] = []
toIndexedCnf n (x:xs) = (Nothing, (n,x)):(toIndexedCnf (n+1) xs)



{- |For implementing a timeout, thanks to HaskellWiki-}
takeFirst :: Int -> IO a -> IO ()
takeFirst 0 action = action >> return ()
takeFirst time action = do  
                            mvar <- newEmptyMVar
                            let sectomicro = 1000000
                                act2 :: IO (Maybe a)
                                act2 = threadDelay (time*sectomicro) >> return Nothing
                                f action = forkIO $ action >>= (putMVar mvar)
                            computations <- mapM f [fmap Just action, act2]
                            result <- takeMVar mvar
                            mapM_ killThread computations
                            case result of
                                 Nothing -> putStrLn $ "Timing out....." ++ show time ++ " seconds over."
                                 otherwise -> return ()

{- IO Functions follow -}

{- A wrapper around readline -}
getline :: String -> IO String
getline s = do 
	x <- readline s
	case x of
             Just y ->  do
		     	addHistory y
	     		return y
             Nothing -> error "\nBye!!"


getClauses :: IO [CnfFormula]
getClauses = do --putStrLn "Enter File name: "
                x <- getline "Enter File name: "
                Exception.catch (getFromFile x)
                      (\e -> do
			   let err = show (e :: Exception.Exception)   
                           hPutStr stderr ("Error with file:" ++ x ++ " : " ++ err ++ " Try again!\n")
                           getClauses)
                             

getTimeOut :: IO Int
getTimeOut = do --putStrLn "Enter the timeout(in seconds, 0 for no timeout): "
                time <- getline "Enter the timeout(in seconds, 0 for no timeout): "
                let x = parse Resolver1.int "input" time
                case x of
                     Left e -> do hPutStrLn stderr $ show e
                                  getTimeOut
                     Right r -> return r

getSupSuggestion :: [IndexedClause] -> IO ([IndexedClause], [IndexedClause])
getSupSuggestion list = do 
        suggestion <- getSupSuggestion'
        if suggestion == []
           then return ([head(list)], tail list)
           else return $ partition (f suggestion) list
                        where 
                                f :: [Int] -> IndexedClause -> Bool
                                f suggestion (_, (a,_)) = elem a suggestion
                                getSupSuggestion' :: IO [Int]        
                                getSupSuggestion' = do
                                        --putStrLn "Enter a suggestion for the set of support([] for default):" 
                                        list <- getline "Enter a suggestion for the set of support([] for default):" 
                                        let x = parse intList "input" list
                                        case x of
                                             Left e -> do hPutStrLn stderr $ show e
                                                          getSupSuggestion'
                                             Right r -> return r

welcome :: IO()
welcome = do 
                putStrLn "\nATP version 1.0\n"
                putStrLn "Starting new session...\n"
                putStrLn "To Exit midway, enter EOF (Ctrl-D on *nix systems)."

getfile:: IO ()
getfile = do 
             list <- getClauses
             let iList = List.reverse $ toIndexedCnf 0 list
             case iList of
                  [] -> putStrLn "Empty input. Possibly the file is incorrectly formatted?" >> getfile    
                  otherwise -> do 
                                   let (support:kb) = iList
                                   putStrLn $ showIndexedList kb
                                   putStrLn $ "\nSupport: \n" ++ showIndexed support ++ "\n"
                                   (support1, kb1) <- getSupSuggestion iList
                                   time <- getTimeOut
                                   takeFirst time $ repeatedResolver (List.length iList) support1 kb1
                                   putStrLn "Starting new session...\n"
                                   getfile


